#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <unistd.h>
#include <syslog.h>
#include "libxml/parser.h"
#include "libxml/tree.h"
#include "libpq-fe.h"
#include "reporting_config.h"
#include <openssl/rc4.h>
#include "inno-lib.h"

#define ERROR_SIZE      4096
#define CONFIG_FILE     "/home/innovaphone-reporting/reporting.conf"

char * copyPQValue(char * value)
{
    char * v = NULL;
    if(value) {
        v = str_malloc(value);
    }
    else {
        v = str_malloc("");
    }
    return v;
}

void set_error(config_t * cfg, char * format, ...)
{
    if(cfg->error) free(cfg->error);
    cfg->error = malloc(ERROR_SIZE);

    va_list args;
    va_start (args, format);
    vsnprintf (cfg->error, ERROR_SIZE, format, args);
    va_end (args);
}

void free_ldap_config(ldap_servers_t *ldap_servers)
{
    if(ldap_servers) {
        if(trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "free ldap config %x %d", ldap_servers, ldap_servers->num_ldap_servers);
        }
        if(ldap_servers->ldap) {
            ldap_t **ldap_tmp = ldap_servers->ldap;

            while(*ldap_tmp != NULL) {
                ldap_t *ldap_cfg = *ldap_tmp++;
                int i;

                if(ldap_cfg->pbx) free(ldap_cfg->pbx);
                if(ldap_cfg->server) free(ldap_cfg->server);
                if(ldap_cfg->host) free(ldap_cfg->host);
                if(ldap_cfg->port) free(ldap_cfg->port);
                if(ldap_cfg->user) free(ldap_cfg->user);
                if(ldap_cfg->password) free(ldap_cfg->password);
                if(ldap_cfg->base) free(ldap_cfg->base);
                if(ldap_cfg->filter) free(ldap_cfg->filter);
                for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
                    if(ldap_cfg->nameattrib[i] != NULL) free(ldap_cfg->nameattrib[i]);
                }
                for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
                    if(ldap_cfg->numberattrib[i] != NULL) free(ldap_cfg->numberattrib[i]);
                }
                if(ldap_cfg->active) free(ldap_cfg->active);
                if(ldap_cfg->cc) free(ldap_cfg->cc);
                if(ldap_cfg->ac) free(ldap_cfg->ac);
                if(ldap_cfg->np) free(ldap_cfg->np);
                if(ldap_cfg->ip) free(ldap_cfg->ip);
                if(ldap_cfg->el) free(ldap_cfg->el);
                if(ldap_cfg->sn) free(ldap_cfg->sn);
                 
                free(ldap_cfg);
            }
            free(ldap_servers->ldap);
        }
        free(ldap_servers);
    }
}

void cleanup_config(config_t * cfg)
{
    if(cfg->error) free(cfg->error);
    if(cfg->replication)
    {
        if(cfg->replication->server) free(cfg->replication->server);
        if(cfg->replication->mode) free(cfg->replication->mode);
        if(cfg->replication->password) free(cfg->replication->password);
        if(cfg->replication->interval) free(cfg->replication->interval);
        free(cfg->replication);
    }
    if(cfg->pbx) 
    {
        if(cfg->pbx->identifier) free(cfg->pbx->identifier);
    }

    free_ldap_config(cfg->ldap_servers);

    free(cfg);
}

ldap_servers_t * read_ldap_config(PGconn *conn, pthread_mutex_t *mutex, ldap_servers_t **ldap_act)
{
    int num_servers = 0, i = 0, j = 0;
    PGresult *res = NULL;
    char * attrib = NULL;
    ldap_servers_t *ldap_servers = NULL;
    int remove_file = 0;
    FILE * fp;
    PGconn *ldap_conn = NULL;

    if(mutex != NULL) {   // cdr
        if(trace_level & TRACE_STD) {	     
            do_log(LOG_REPORTING, "", "Ldap lock %s", LDAP_SERVER_FILE);
        }	     
        pthread_mutex_lock(mutex);
        if((fp = fopen(LDAP_SERVER_FILE, "r")) != NULL) {
            fclose(fp);

            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "Ldap file changed!");
            }

            ldap_conn = PQconnectdb(PG_CONN_STR);
            if(PQstatus(ldap_conn) != CONNECTION_OK) {
                do_log(LOG_REPORTING, "", "Ldap: Connection to database failed: %s", PG_CONN_STR);
                PQfinish(ldap_conn);
                pthread_mutex_unlock(mutex);
                return NULL;
            }
            if(PQsetClientEncoding(ldap_conn, "UTF-8") != 0) {
                do_log(LOG_REPORTING, "", "Ldap: Setting client encoding failed: %s", "UTF-8");
                PQfinish(ldap_conn);
                pthread_mutex_unlock(mutex);
                return NULL;
            }
            res = PQexec(ldap_conn, PG_LDAP_ACT_QUERY);
            if(PQresultStatus(res) != PGRES_TUPLES_OK) {
                do_log(LOG_REPORTING, "", "Ldap: query error: %s", PQerrorMessage(ldap_conn));
                PQclear(res);
                PQfinish(ldap_conn);
                pthread_mutex_unlock(mutex);
                return NULL;
            }
            num_servers = PQntuples(res);
            remove_file = 1;
        }
        else {
            pthread_mutex_unlock(mutex);

            return NULL;
        }
    }
    else if(conn != NULL) {
        res = PQexec(conn, PG_LDAP_ALL_QUERY);
        if(PQresultStatus(res) != PGRES_TUPLES_OK) {
            do_log(LOG_REPORTING, "", "reporting ldap: query error: %s", PQerrorMessage(conn));
            PQclear(res);
            return NULL;
        }
        num_servers = PQntuples(res);
    }

    if((ldap_servers = (ldap_servers_t *)malloc(sizeof(ldap_servers_t))) == NULL) {
        do_log(LOG_REPORTING, "", "No memory for ldap server config!");
        if(mutex) pthread_mutex_unlock(mutex);
        return NULL;
    }
    memset(ldap_servers, 0, sizeof(ldap_servers_t));
    if(num_servers) {
        ldap_servers->num_ldap_servers = num_servers;
        if((ldap_servers->ldap = (ldap_t **)malloc((num_servers+1)*sizeof(ldap_t *))) == NULL) {
            do_log(LOG_REPORTING, "", "No memory for ldap servers!");
            free(ldap_servers);
            if(mutex) pthread_mutex_unlock(mutex);
            return NULL;
        }
        memset(ldap_servers->ldap, 0, (num_servers+1)*sizeof(ldap_t *));

        ldap_t **ldap_tmp = ldap_servers->ldap;
        ldap_t *ldap_cfg;
        for (i = 0; i < num_servers; i++) {
            *ldap_tmp = (ldap_t *)malloc(sizeof(ldap_t));
            ldap_cfg = *ldap_tmp++;
		    memset(ldap_cfg, 0x0, sizeof(ldap_t));

            ldap_cfg->server = copyPQValue(PQgetvalue(res, i, 0));
            ldap_cfg->host = copyPQValue(PQgetvalue(res, i, 1));
            ldap_cfg->port = copyPQValue(PQgetvalue(res, i, 2));
            ldap_cfg->password = copyPQValue(PQgetvalue(res, i, 3));
            ldap_cfg->user = copyPQValue(PQgetvalue(res, i, 4));
            ldap_cfg->base = copyPQValue(PQgetvalue(res, i, 5));
            ldap_cfg->filter = copyPQValue(PQgetvalue(res, i, 16));
            attrib = strtok (PQgetvalue(res, i, 6), ",");
            if(attrib == NULL) do_log(LOG_REPORTING, "", "Number attributes are NULL!");
            j = 0;
            while (attrib != NULL && j < MAX_NUM_ATTRIBUTES)
            {
                ldap_cfg->numberattrib[j] = str_malloc(attrib);
                attrib = strtok (NULL, ",");
                j++;
            }
            attrib = strtok (PQgetvalue(res, i, 7), ",");
            if(attrib == NULL) do_log(LOG_REPORTING, "", "Name attributes are NULL!");
            j = 0;
            while (attrib != NULL && j < MAX_NUM_ATTRIBUTES)
            {
                ldap_cfg->nameattrib[j] = str_malloc(attrib);
                attrib = strtok (NULL, ",");
                j++;
            }
            ldap_cfg->pbx = copyPQValue(PQgetvalue(res, i, 8));
            ldap_cfg->cc = copyPQValue(PQgetvalue(res, i, 9));
            ldap_cfg->ac = copyPQValue(PQgetvalue(res, i, 10));
            ldap_cfg->np = copyPQValue(PQgetvalue(res, i, 11));
            ldap_cfg->ip = copyPQValue(PQgetvalue(res, i, 12));
            ldap_cfg->el = copyPQValue(PQgetvalue(res, i, 13));
            ldap_cfg->sn = copyPQValue(PQgetvalue(res, i, 14));
            // active TRUE is 't', FALSE is 'f'
            ldap_cfg->active = copyPQValue(PQgetvalue(res, i, 15));

            if(ldap_cfg->password && *ldap_cfg->password &&\
               ldap_cfg->host && *ldap_cfg->host) {
                char * decrypted = rc4_decrypt(ldap_cfg->password, ldap_cfg->host);
                free(ldap_cfg->password);
                ldap_cfg->password = decrypted;
            }
        }
    }
    if(mutex) {
        PQclear(res);
        PQfinish(ldap_conn);
        if(remove_file) {
            unlink(LDAP_SERVER_FILE);
        }
        free_ldap_config(*ldap_act);
        *ldap_act = ldap_servers;
        pthread_mutex_unlock(mutex);
        return NULL; 
    }
    else if(conn != NULL) {
        PQclear(res);
    }	     
    return ldap_servers;
}

config_t * read_config(PGconn * conn) 
{
    xmlDoc *doc = NULL;
    xmlNodePtr root = NULL;
    xmlNodePtr node = NULL;
    xmlNodePtr child = NULL;
    
    config_t * cfg = malloc(sizeof(config_t));
    memset(cfg, 0, sizeof(config_t));

    if ((doc = xmlReadFile(CONFIG_FILE, NULL, 0)) == NULL) {
        set_error(cfg, "Couldn't read reporting config file '%s'.", CONFIG_FILE);
        return cfg;
    } 

    root = xmlDocGetRootElement(doc);
    if(root == NULL) {
        set_error(cfg, "No root element found in reporting config file '%s'.", CONFIG_FILE);
        return cfg;
    }

    cfg->ldap_servers = read_ldap_config(conn, NULL, NULL);

    cfg->replication = malloc(sizeof(replication_t));
    memset(cfg->replication, 0, sizeof(replication_t));
    
    cfg->pbx = malloc(sizeof(pbx_t));
    memset(cfg->pbx, 0, sizeof(pbx_t));

    for(node = root->children; node != NULL; node = node->next)
    {
        if(!xmlStrcasecmp(node->name, (xmlChar *)"replication")) {
            for(child = node->children; child != NULL; child = child->next) 
            {
                if(!xmlStrcasecmp(child->name, (xmlChar *)"server")) {
                    cfg->replication->server = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"mode")) {
                    cfg->replication->mode = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"password")) {
                    cfg->replication->password = xmlNodeGetContentChar(child);
                }
                else if(!xmlStrcasecmp(child->name, (xmlChar *)"interval")) {
                    cfg->replication->interval = xmlNodeGetContentChar(child);
                }
            }
        }
        if(!xmlStrcasecmp(node->name, (xmlChar *)"pbx")) {
            for(child = node->children; child != NULL; child = child->next) 
            {
                if(!xmlStrcasecmp(child->name, (xmlChar *)"identifier")) {
                    cfg->pbx->identifier = xmlNodeGetContentChar(child);
                }
            }
        }
    }
    if(!cfg->pbx->identifier) { // set default if config doesn't contain identifier
        cfg->pbx->identifier = str_malloc("sys");
    }

    xmlFreeDoc(doc);
    doc = NULL;

    if(cfg->replication->password && *cfg->replication->password &&\
       cfg->replication->server && *cfg->replication->server) {
        char * decrypted = rc4_decrypt(cfg->replication->password, cfg->replication->server);
        free(cfg->replication->password);
        cfg->replication->password = decrypted;
    }

    return cfg;
}
